# bodhi documentation build configuration file, created by
# sphinx-quickstart on Sat Aug 10 09:29:50 2013.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.
import datetime

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'bodhi'
copyright = '2007-{}, Red Hat, Inc.'.format(datetime.datetime.utcnow().year)

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('man_pages/bodhi', 'bodhi', 'manage Fedora updates', ['Randy Barlow', 'Luke Macken'], 1),
]
